from django.test import TestCase
from django.test.utils import override_settings
from django.test.client import RequestFactory

from django.core.urlresolvers import reverse
from django.template import Template, Context

from django.contrib.auth.models import AnonymousUser

from django_dynamic_fixture import N, G

from django_webtest import WebTest
from webtest.response import TestResponse

from ..models import WishlistItem
from ..utils import get_user_model
User = get_user_model()

from ..context_processors import wishlist_items

from .models import TestItemModel


@override_settings(WISHLIST_ITEM_MODEL='tests.TestItemModel')
class WishlistUnitTests(TestCase):
    """ Unit tests for wishlists. """

    def test_save(self):
        """ Test saving a WishlistItem. """

        # Create
        item = N(WishlistItem)

        # Validate
        item.clean()

        # Save
        item.save()

    def test_unicode(self):
        """ Test __unicode__ """
        item = G(WishlistItem)

        self.assertEquals(unicode(item), unicode(item.item))

    def test_absolute_url(self):
        """ Test get_absolute_url() """
        item = G(WishlistItem)

        self.assertEquals(
            item.get_absolute_url(),
            item.item.get_absolute_url()
        )

    def test_manager(self):
        """ Test custom user manager """

        # Create two users and wishlist items
        user1 = G(User)
        user2 = G(User)

        item1 = G(WishlistItem, user=user1)
        item2 = G(WishlistItem, user=user2)

        # Only this item should be available
        self.assertEquals(
            WishlistItem.objects.for_user(user=user1).count(),
            1
        )

        self.assertEquals(
            WishlistItem.objects.for_user(user=user1)[0],
            item1
        )

        self.assertEquals(
            WishlistItem.objects.for_user(user=user2).count(),
            1
        )

        self.assertEquals(
            WishlistItem.objects.for_user(user=user2)[0],
            item2
        )

    def test_context_processor(self):
        """ Test wishlist_items context processor. """

        # Create one item for testing
        item = G(WishlistItem)

        # Request factory and request
        factory = RequestFactory()
        request = factory.get('/')

        # Anonymous user
        request.user = AnonymousUser()

        # No context for anonymous users
        result = wishlist_items(request)
        self.assertEquals(result, {})

        # User set, should return items for user
        request.user = item.user
        result = wishlist_items(request)

        self.assertEquals(len(result['wishlist_items']), 1)

        self.assertEquals(
            result['wishlist_items'][0],
            WishlistItem.objects.for_user(user=item.user)[0]
        )

    def test_add_form(self):
        """ Test wishlist_add_form template tag. """

        # Create one wished item for testing
        wished_item = G(TestItemModel)

        # Get URL for add view
        add_view = reverse('wishlist_add')

        # Create user for testing
        user = G(User)

        # Setup and render template
        template = Template(
            '{% load wishlist_tags %}{% wishlist_add_form item %}'
        )
        context = Context(dict(item=wished_item, user=user))
        rendered = template.render(context)

        # Test output
        self.assertHTMLEqual(
            '<form action="{0}" method="post">'
            '<input type="hidden" name="item" value="{1}">'
            '<input type="submit" value="Add">'
            '</form>'.format(add_view, wished_item.pk),
            rendered
        )

    def test_remove_form(self):
        """ Test wishlist_remove_form template tag. """

        # Create one item for testing
        wishlist_item = G(WishlistItem)

        # Remove url
        remove_view = reverse(
            'wishlist_remove', kwargs=dict(pk=wishlist_item.pk)
        )

        # Setup and render template
        template = Template(
            '{% load wishlist_tags %}{% wishlist_remove_form item %}'
        )
        context = Context(
            dict(item=wishlist_item.item, user=wishlist_item.user)
        )
        rendered = template.render(context)

        # Test output
        self.assertHTMLEqual(
            '<form action="{0}" method="post">'
            '<input type="submit" value="Remove">'
            '</form>'.format(
                remove_view
            ), rendered
        )


@override_settings(WISHLIST_ITEM_MODEL='tests.TestItemModel')
class WishlistFunctionalTests(WebTest):
    """ Functional/integration tests for views. """

    def test_login_required(self):
        """ Assure login is required. """

        view = reverse('wishlist_add')

        result = self.app.get(view)

        self.assertEquals(result.status_int, 302)
        self.assertIn('login', result.url)

    def test_add(self):
        """ Test adding an item. """

        # Get URL for add view
        add_view = reverse('wishlist_add')
        list_view = reverse('wishlist')

        # Create an item to add
        wished_item = G(TestItemModel, slug='test')

        # Create a user and login
        user = G(User)

        # Request add view
        add_page = self.app.get(add_view, user=user)

        # Fill in form and post
        add_form = add_page.form
        add_form['item'] = wished_item.pk
        add_result = add_form.submit()

        # Assert WishlistItem has been created
        self.assertEquals(WishlistItem.objects.count(), 1)

        item = WishlistItem.objects.get()
        self.assertEquals(item.user, user)
        self.assertEquals(item.item_id, wished_item.pk)

        # Assert redirect location
        self.assertTrue(add_result.location.endswith(list_view))

        # Test redirect after adding
        result = add_result.follow()

        # Test messages after adding
        self.assertContains(result, unicode(item))
        self.assertContains(result, u'added to the wishlist')

    def test_add_twice(self):
        """
        Test adding an item twice.

        Should add a warning message and not generate an exception.
        """

        # Get URL for add view
        add_view = reverse('wishlist_add')
        list_view = reverse('wishlist')

        # Create an item to add
        wished_item = G(TestItemModel, slug='test')

        # Create a user and login
        user = G(User)

        # Request add view
        add_page = self.app.get(add_view, user=user)

        # Fill in form and post
        add_form = add_page.form
        add_form['item'] = wished_item.pk
        add_result = add_form.submit()

        # Assert WishlistItem has been created
        self.assertEquals(WishlistItem.objects.count(), 1)

        # Test redirect after adding
        result = add_result.follow()
        self.assertContains(result, u'added to the wishlist')

        # Post again
        add_result = add_form.submit()

        # Assert WishlistItem still in wishlist
        self.assertEquals(WishlistItem.objects.count(), 1)

        # Assert redirect location
        self.assertTrue(add_result.location.endswith(list_view))

        # Test redirect after adding
        result = add_result.follow()

        # Test redirect after adding
        self.assertContains(result, u'already in the wishlist')

    def test_remove_confirm(self):
        """ Test removal confirm form. """

        # Create wishlist item
        item = G(WishlistItem)

        # Get URL for remove view
        remove_view = reverse('wishlist_remove', kwargs=dict(pk=item.pk))

        # Request view
        page = self.app.get(remove_view, user=item.user)
        self.assertContains(page, 'Confirm')

        # Post form
        page.form.submit()

        # Assert WishlistItem has been removed
        self.assertEquals(WishlistItem.objects.count(), 0)

    def test_list(self):
        """ Test list view. """

        # Create wishlist item
        item = G(WishlistItem)

        # Get URL for list view
        list_view = reverse('wishlist')

        # Request view
        list_page = self.app.get(list_view, user=item.user)

        self.assertContains(list_page, unicode(item))
        self.assertContains(list_page, item.get_absolute_url())
        self.assertContains(list_page, '1 item in wishlist')

        # Add another item for same user
        item2 = G(WishlistItem, user=item.user)

        # Request view
        list_page = self.app.get(list_view, user=item.user)

        self.assertContains(list_page, unicode(item))
        self.assertContains(list_page, item.get_absolute_url())
        self.assertContains(list_page, unicode(item2))
        self.assertContains(list_page, item2.get_absolute_url())
        self.assertContains(list_page, '2 items in wishlist')

    def test_remove(self):
        """ Test removing an item. """

        # Create wishlist item
        item = G(WishlistItem)

        # Get URL for list view
        list_view = reverse('wishlist')

        # Request view
        list_page = self.app.get(list_view, user=item.user)

        # Find and post remove form
        remove_form = list_page.form
        remove_result = remove_form.submit()

        # Assert WishlistItem has been removed
        self.assertEquals(WishlistItem.objects.count(), 0)

        # Perform redirect after adding
        result = remove_result.follow()

        # Test messages after adding
        self.assertContains(result, unicode(item))
        self.assertContains(result, u'removed from the wishlist')

    def test_permission(self):
        """ List for one user should be empty for another. """

        # Create wishlist item
        item = G(WishlistItem)

        # Get URL for list view
        list_view = reverse('wishlist')

        # Request view
        list_page = self.app.get(list_view, user=item.user)

        # Item for creating user
        self.assertContains(list_page, unicode(item))

        # Request view with another user
        list_page = self.app.get(list_view, user=G(User))

        # No item for other user
        self.assertNotContains(list_page, unicode(item))

    def test_clear(self):
        """ Test clearing the wishlist. """

        # Begin at list view with single item
        # Create wishlist item
        item = G(WishlistItem)

        # Get URL for list view
        list_view = reverse('wishlist')

        # Request view
        list_page = self.app.get(list_view, user=item.user)

        # Click clear button
        clear_page = list_page.click('Clear wishlist')
        self.assertContains(clear_page, 'Confirm')

        # Post form
        clear_page.form.submit()

        # Assert WishlistItem has been removed
        self.assertEquals(WishlistItem.objects.count(), 0)



from .utils import SettingsBase


class WishlistSettings(SettingsBase):
    """ Settings specific to wishlists. """
    settings_prefix = 'WISHLIST'

    DEFAULT_REDIRECT_URL = 'wishlist'

wishlist_settings = WishlistSettings()



from .models import WishlistItem


def wishlist_items(request):
    """ Add wishlist items to request context. """

    if request.user.is_authenticated():
        return {
            'wishlist_items': WishlistItem.objects.for_user(user=request.user)
        }

    # No user, no wishlist
    return {}



from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError

from django.views.generic import ListView, CreateView, DeleteView
from django.http import HttpResponseRedirect

from django.contrib import messages
from django.contrib.auth.decorators import login_required

from django.utils.http import is_safe_url
from django.utils.decorators import method_decorator

from .settings import wishlist_settings
from .models import WishlistItem
from .forms import WishlistItemForm


class WishlistViewMixin(object):
    """ Common mixin for WishlistItem. """

    model = WishlistItem
    form_class = WishlistItemForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        """ Require user login. """

        return super(WishlistViewMixin, self).dispatch(*args, **kwargs)

    def get_queryset(self):
        """ Return only items for current user. """

        user = self.request.user
        assert user.is_authenticated(), \
            'View should not be accesible for unauthenticated users.'

        qs = super(WishlistViewMixin, self).get_queryset()

        return qs.for_user(user=user)

    def get_success_url(self):
        """
        Allow specification of return URL using hidden form field or GET
        parameter, similar to auth views.

        1. Use 'next' POST parameter if available.
        2. Use 'next' GET parameter if available.
        3. Fall back to 'REDIRECT_URL' named URL, defaulting
           to 'wishlist' view.
        """

        redirect_to = self.request.POST.get(
            'next', self.request.GET.get('next', '')
        )

        if not is_safe_url(url=redirect_to, host=self.request.get_host()):
            redirect_to = reverse(wishlist_settings.REDIRECT_URL)

        return redirect_to


class WishlistView(WishlistViewMixin, ListView):
    """ View providing list if WishlistItem's for User. """

    pass


class WishlistClearView(WishlistViewMixin, ListView):
    """ Clear the wishlist. """

    template_name_suffix = '_confirm_clear'

    def post(self, *args, **kwargs):
        """ Delete and report to user. """

        # Delete all items
        self.get_queryset().delete()

        # Create message
        messages.add_message(
            self.request,
            messages.SUCCESS,
            u'The wishlist has been cleared.'
        )

        # Return result of super
        return HttpResponseRedirect(self.get_success_url())


class WishlistAddView(WishlistViewMixin, CreateView):
    """ Add item to the wishlist. """

    def form_valid(self, form):
        """ Save and report to user. """

        # Set user
        form.instance.user = self.request.user

        # Get created instance
        self.object = form.save(commit=False)

        # Validate uniqueness
        try:
            form.instance.validate_unique()

        except ValidationError:
            messages.add_message(
                self.request, messages.ERROR,
                u'Item {item} is already in the wishlist.'.format(
                    item=unicode(form.instance.item)
                )
            )

            # Return redirect to success URL
            return HttpResponseRedirect(self.get_success_url())

        # Call super
        result = super(WishlistAddView, self).form_valid(form)

        # Create message
        messages.add_message(
            self.request, messages.SUCCESS,
            u'Item {item} has been added to the wishlist.'.format(
                item=unicode(form.instance.item)
            )
        )

        # Return super
        return result


class WishlistRemoveView(WishlistViewMixin, DeleteView):
    """ Delete item from the wishlist. """

    def delete(self, request, *args, **kwargs):
        """ Delete and report to user. """

        # Call super
        result = super(WishlistRemoveView, self).delete(
            self, request, *args, **kwargs
        )

        # Create message
        messages.add_message(
            self.request,
            messages.SUCCESS,
            'Item {item} has been removed from the wishlist.'.format(
                item=unicode(self.object)
            )
        )

        # Return result of super
        return result



from django.db import models


class TestItemModel(models.Model):
    """ Item used for testing. """

    slug = models.SlugField()

    def __unicode__(self):
        return self.slug

    def get_absolute_url(self):
        return 'bogus'



from django.forms import ModelForm

from .models import WishlistItem


class WishlistItemForm(ModelForm):
    """ Form for WishlistItem. """

    class Meta:
        model = WishlistItem
        fields = ['item']



from django.db.models.query import QuerySet
from django.db.models import Manager


class UserQuerySet(QuerySet):
    """ Custom QuerySet for UserManager. """

    def for_user(self, user):
        """ Return items for a given user. """
        return self.filter(user=user)


class UserManager(Manager):
    """ Custom manager adding for_user(user) method. """

    def get_queryset(self):
        """ Return a UserQuerySet instead of QuerySet. """
        return UserQuerySet(self.model, using=self._db)

    def for_user(self, user):
        """ Wrapper around for_user() QS method. """
        return self.get_queryset().for_user(user=user)

    def get_query_set(self):
        """ Django <1.6 compatibility wrapper. """
        return self.get_queryset()



from django.contrib import admin


from .models import WishlistItem


class WishlistItemAdmin(admin.ModelAdmin):
    """ Admin for WishlistItem. """

    pass

admin.site.register(WishlistItem, WishlistItemAdmin)



from django import template

from copy import copy

register = template.Library()


def wishlist_form(context, item):
    """
    Render wishlist form for adding or removing a particular item.

    Assumes the user is available in the context and requires an instance of
    the wishlist model *not* WishlistItem.
    """

    # TODO: Replace by proper, meaningful, exceptions.
    assert 'user' in context
    assert hasattr(item, 'pk')
    assert hasattr(item, 'wishlistitem_set')

    # Copy local context dict, including
    context = copy(context)

    # Add item
    context['item'] = item

    # Return context for use in inclusion tag
    return context

# Register for adding
register.inclusion_tag(
    'wishlist/wishlistitem_add_form.html',
    takes_context=True, name='wishlist_add_form'
)(wishlist_form)

# Register for removing
register.inclusion_tag(
    'wishlist/wishlistitem_remove_form.html',
    takes_context=True, name='wishlist_remove_form'
)(wishlist_form)



from django.conf import settings as django_settings
from django.utils.importlib import import_module


class Singleton(type):
    """
    Singleton metaclass.
    Source:
    http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs
            )

        return cls._instances[cls]


class SettingsBase(object):
    """
    A settings object that proxies settings and handles defaults, inspired
    by `django-appconf` and the way it works  in `django-rest-framework`.

    By default, a single instance of this class is created as `<app>_settings`,
    from which `<APP>_SETTING_NAME` can be accessed as `SETTING_NAME`, i.e.::

        from myapp.settings import myapp_settings

        if myapp_settings.SETTING_NAME:
            # DO FUNKY DANCE

    If a setting has not been explicitly defined in Django's settings, defaults
    can be specified as `DEFAULT_SETTING_NAME` class variable or property.
    """

    __metaclass__ = Singleton

    def __init__(self):
        """
        Assert app-specific prefix.
        """
        assert hasattr(self, 'settings_prefix'), 'No prefix specified.'

    def __getattr__(self, attr):
        """
        Return Django setting `PREFIX_SETTING` if explicitly specified,
        otherwise return `PREFIX_SETTING_DEFAULT` if specified.
        """

        if attr.isupper():
            # Require settings to have uppercase characters

            try:
                setting = getattr(
                    django_settings,
                    '%s_%s' % (self.settings_prefix, attr),
                )
            except AttributeError:
                if not attr.startswith('DEFAULT_'):
                    setting = getattr(self, 'DEFAULT_%s' % attr)
                else:
                    raise

            return setting

        else:
            # Default behaviour
            raise AttributeError(
                'No setting or default available for \'%s\'' % attr
            )


def import_object(from_path):
    """ Given an import path, return the object it represents. """

    module, attr = from_path.rsplit(".", 1)
    mod = import_module(module)
    return getattr(mod, attr)


def get_user_model():
    """ get_user_model compatibility wrapper. Returns active User model. """
    try:
        from django.contrib.auth import get_user_model
    except ImportError:
        # Django < v1.5
        from django.contrib.auth.models import User
    else:
        User = get_user_model()

    return User



#!/usr/bin/env python
# This file is part of django-url-sso.
#
# django-url-sso: Generate login URL's for unstandardized SSO systems.
# Copyright (C) 2014 Mathijs de Bruin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

try:
    README = open('README.rst').read()
except:
    README = None

try:
    REQUIREMENTS = open('requirements.txt').read()
except:
    REQUIREMENTS = None

setup(
    name='django-wishlist',
    version='0.1',
    description='Generic user wishlists for use with any Django model.',
    long_description=README,
    install_requires=REQUIREMENTS,
    license='AGPL',
    author='Mathijs de Bruin',
    author_email='mathijs@mathijsfietst.nl',
    url='https://github.com/dokterbob/django-wishlist/',
    packages=find_packages(),
    include_package_data=True,
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities'
    )
)



from django.conf.urls import patterns, url

from .views import (
    WishlistView, WishlistAddView, WishlistClearView, WishlistRemoveView
)


urlpatterns = patterns('',
    url(r'^$',
        WishlistView.as_view(), name='wishlist'),
    url(r'^add/$',
        WishlistAddView.as_view(), name='wishlist_add'),
    url(r'^clear/$',
        WishlistClearView.as_view(), name='wishlist_clear'),
    url(r'^(?P<pk>\d+)/remove/$',
        WishlistRemoveView.as_view(), name='wishlist_remove'),
)



from django.db import models

from django.utils.translation import ugettext as _

from .settings import wishlist_settings

from .managers import UserManager

from .utils import get_user_model
User = get_user_model()


class WishlistItem(models.Model):
    """ Item in wishlist. """

    created = models.DateTimeField(_('created'), auto_now_add=True)

    user = models.ForeignKey(User)
    item = models.ForeignKey(wishlist_settings.ITEM_MODEL)

    objects = UserManager()

    class Meta:
        verbose_name = _('wishlist')
        verbose_name_plural = _('wishlists')
        ordering = ['-created']
        unique_together = ['user', 'item']

    def __unicode__(self):
        """ Wrap unicode of item. """

        assert self.item

        return unicode(self.item)

    def get_absolute_url(self):
        """ Return absolute URL for item. """

        return self.item.get_absolute_url()
